/* 
 *	reads current (>3.1 or so?) GMT grid, writes older style (<3.1).
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <netcdf.h>
#include <gmt.h>

#ifndef TRUE
#define TRUE  1
#define FALSE 0
#endif

int GMT_cdf_read_grd_info ();
int GMT_cdf_read_grd ();

int write_grd ();
char *memory();

/* some things I probably shouldn't do */

#define MAXDOUBLE 1.0e32
#define bad_float(x) ((x) != (x))


main (argc, argv)
char *argv[];
{
     int		numPoints;
     float		*gmtValue;
     struct GRD_HEADER 	gmtGrd;
     char 		inDatafile[128];
     char		outDatafile[128];


     /* check proper arguments */

     if (argc < 3)  {
	  printf("Usage: grdcvt inputfile outputfile\n");
	  return FALSE;
     }

     /* open and read the GMT Grid Header */

     strcpy(inDatafile,argv[1]);
     if (GMT_cdf_read_grd_info (inDatafile, &gmtGrd)) {
	  printf ("Error reading GMT grid header information\n");
	  return FALSE;
     }

     /* and actual data too */

     numPoints = gmtGrd.nx * gmtGrd.ny;
     gmtValue = (float *) malloc (sizeof(float) * numPoints);
     if (GMT_cdf_read_grd (inDatafile, &gmtGrd, gmtValue,
			   0.0,0.0,0.0,0.0,GMT_pad,FALSE)) {
	  printf ("Error reading GMT grid data\n");
	  return FALSE;
     }

     /* now just write */

     strcpy(outDatafile,argv[2]);
     if (write_grd(outDatafile,&gmtGrd,gmtValue,
		   0.0,0.0,0.0,0.0,GMT_pad,FALSE)) {
	  printf("Error writing grid\n");
	  return FALSE;
     }     


     /* all done */

     free (gmtValue);
}        

/*
 * stolen from GMT3.0/src/gmt_cdf.c :
 */

/*--------------------------------------------------------------------
 *    The GMT-system:	@(#)gmt_cdf.c	2.19  2/23/95
 *
 *    Copyright (c) 1991-1995 by P. Wessel and W. H. F. Smith
 *    See README file for copying and redistribution conditions.
 *--------------------------------------------------------------------*/
/*
 *
 *	G M T _ C D F . C   R O U T I N E S
 *
 * Takes care of all grd input/output built on NCAR's netCDF routines (which is
 * an XDR implementation)
 * Most functions return the ncerr error value which will be non-zero if
 * an error occured.
 *
 * Author:	Paul Wessel
 * Date:	9-SEP-1992
 * Modified:	27-JUN-1995
 * Version:	3.0
 *
 * Functions include:
 *
 *	cdf_read_grd_info :	Read header from file
 *	cdf_read_grd :		Read header and data set from file
 *	cdf_write_grd_info :	Update header in existing file
 *	cdf_write_grd :		Write header and data set to new file
 *
 * - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*/


int write_grd (file, header, grid, w, e, s, n, pad, complex)
char *file;
struct GRD_HEADER *header;
float *grid;
double w, e, s, n;	/* Sub-region to write out */
int pad[];		/* padding on w, e, s, n of grid, respectively */
BOOLEAN complex;	/* TRUE if array is to hold real and imaginary parts */
{
	nclong start[2], edge[2];
	nclong i, i2, cdfid, nm[2], inc = 1, *k;
	nclong j, ij, j2, width_in, width_out, height_out, one_or_zero;
	nclong first_col, last_col, first_row, last_row;
	
	BOOLEAN geo = FALSE;
	double dummy[2], off, half_or_zero, small, x;
	
	char text[480];
	
	float *tmp;
	
	/* dimension ids */
	nclong  side_dim, xysize_dim;

	/* variable ids */
	nclong  x_range_id, y_range_id, z_range_id, inc_id, nm_id, z_id;

	/* variable shapes */
	int dims[1];

	if (w == 0.0 && e == 0.0) {	/* Write entire grid */
		w = header->x_min;
		e = header->x_max;
		s = header->y_min;
		n = header->y_max;
		pad[0] = pad[1] = pad[2] = pad[3] = 0;
	}
	
	if (w < header->x_min || e > header->x_max) geo = TRUE;	/* Dealing with periodic grid */
	
	one_or_zero = (header->node_offset) ? 0 : 1;
	off = (header->node_offset) ? 0.0 : 0.5;
	
	/* Create file and enter define mode */
	
	if (ncopts) ncopts = 0;
	if ((cdfid = nccreate (file, NC_CLOBBER)) == -1) return (-1);

	/* Get dimension of subregion to write */
	
	width_out  = rint ((e - w) / header->x_inc) + one_or_zero;
	height_out = rint ((n - s) / header->y_inc) + one_or_zero;
	
	/* Get first and last row and column numbers */
	
	small = 0.1 * header->x_inc;
	first_col = floor ((w - header->x_min + small) / header->x_inc);
	last_col  = ceil ((e - header->x_min - small) / header->x_inc) -1 + one_or_zero;
	small = 0.1 * header->y_inc;
	first_row = floor ((header->y_max - n + small) / header->y_inc);
	last_row  = ceil ((header->y_max - s - small) / header->y_inc) -1 + one_or_zero;
	
	if ((last_col - first_col + 1) > width_out) last_col--;
	if ((last_row - first_row + 1) > height_out) last_row--;
	if ((last_col - first_col + 1) > width_out) first_col++;
	if ((last_row - first_row + 1) > height_out) first_row++;

	width_in = width_out;		/* Physical width of input array */
	if (pad[0] > 0) width_in += pad[0];
	if (pad[1] > 0) width_in += pad[1];
	
	edge[0] = width_out;
	if (complex) inc = 2;

	header->x_min = w;
	header->x_max = e;
	header->y_min = s;
	header->y_max = n;
	
	/* define dimensions */
	
	side_dim	= ncdimdef(cdfid, "side", 2);
	xysize_dim	= ncdimdef(cdfid, "xysize", (int) (width_out * height_out));

	/* define variables */

	dims[0]		= side_dim;
	x_range_id	= ncvardef (cdfid, "x_range", NC_DOUBLE, 1, dims);
	y_range_id	= ncvardef (cdfid, "y_range", NC_DOUBLE, 1, dims);
	z_range_id	= ncvardef (cdfid, "z_range", NC_DOUBLE, 1, dims);
	inc_id		= ncvardef (cdfid, "spacing", NC_DOUBLE, 1, dims);
   	nm_id		= ncvardef (cdfid, "dimension", NC_LONG, 1, dims);

	dims[0]		= xysize_dim;
	z_id		= ncvardef (cdfid, "z", NC_FLOAT, 1, dims);

	/* assign attributes */
	
	strcpy (text, header->command);
	strcpy (&text[320], header->remark);
	ncattput (cdfid, x_range_id, "units", NC_CHAR, 80, (void *)header->x_units);
	ncattput (cdfid, y_range_id, "units", NC_CHAR, 80, (void *)header->y_units);
	ncattput (cdfid, z_range_id, "units", NC_CHAR, 80, (void *)header->z_units);
	ncattput (cdfid, z_id, "long_name", NC_CHAR, 80, (void *)header->title);
	ncattput (cdfid, z_id, "scale_factor", NC_DOUBLE, 1,(void *) &header->z_scale_factor);
	ncattput (cdfid, z_id, "add_offset", NC_DOUBLE, 1,(void *) &header->z_add_offset);
	ncattput (cdfid, z_id, "node_offset", NC_LONG, 1,(void *) &header->node_offset);
	ncattput (cdfid, NC_GLOBAL, "title", NC_CHAR, 80, (void *)header->title);
	ncattput (cdfid, NC_GLOBAL, "source", NC_CHAR, 480, (void *)text);

	/* leave define mode */
	
	ncendef (cdfid);

	/* Find xmin/zmax */
	
	header->z_min = MAXDOUBLE;	header->z_max = -MAXDOUBLE;
	for (j = first_row, j2 = pad[3]; j <= last_row; j++, j2++) {
		for (i = first_col, i2 = pad[0]; i <= last_col; i++, i2++) {
			ij = (j2 * width_in + i2) * inc;
			if (bad_float ((double)grid[ij])) continue;
			header->z_min = MIN (header->z_min, grid[ij]);
			header->z_max = MAX (header->z_max, grid[ij]);
		}
	}
	
	/* store header information and array */
	
	start[0] = 0;
	edge[0] = 2;
	dummy[0] = header->x_min;
	dummy[1] = header->x_max;
	ncvarput(cdfid, x_range_id, start, edge, (void *)dummy);
	dummy[0] = header->y_min;
	dummy[1] = header->y_max;
	ncvarput(cdfid, y_range_id, start, edge, (void *)dummy);
	dummy[0] = header->x_inc;
	dummy[1] = header->y_inc;
	ncvarput(cdfid, inc_id, start, edge, (void *)dummy);
	nm[0] = width_out;
	nm[1] = height_out;
	ncvarput(cdfid, nm_id, start, edge, (void *)nm);
	dummy[0] = header->z_min;
	dummy[1] = header->z_max;
	ncvarput(cdfid, z_range_id, start, edge, (void *)dummy);
	edge[0] = width_out;

	if (geo) {
		tmp = (float *) memory (CNULL, (int)width_in, sizeof (float), "cdf_write_grd");
		k = (nclong *) memory (CNULL, (int)width_out, sizeof (nclong), "cdf_write_grd");
		
		half_or_zero = (header->node_offset) ? 0.5 : 0.0;
		edge[0] = header->nx;
		small = 0.1 * header->x_inc;	/* Anything smaller than 0.5 dx will do */
		for (i = 0; i < width_out; i++) {
			x = w + (i + half_or_zero) * header->x_inc;
			if ((header->x_min - x) > small)
				x += 360.0;
			else if ((x - header->x_max) > small)
				x -= 360.0;
			k[i] = (nclong) floor (((x - header->x_min) / header->x_inc) + off);
		}
		i2 = first_col + pad[0];
		for (j = 0, j2 = first_row + pad[3]; j < height_out; j++, j2++) {
			ij = j2 * width_in + i2;
			start[0] = j * width_out;
			for (i = 0; i < width_out; i++) tmp[i] = grid[inc * (ij+k[i])];
			ncvarput (cdfid, z_id, start, edge, (void *)tmp);
		}
		free ((char *)k);
	}
	else {
		if (complex) tmp = (float *) memory (CNULL, (int)width_in, sizeof (float), "cdf_write_grd");
		i2 = first_col + pad[0];
		for (j = 0, j2 = first_row + pad[3]; j < height_out; j++, j2++) {
			ij = j2 * width_in + i2;
			start[0] = j * width_out;
			if (complex) {
				for (i = 0; i < width_out; i++) tmp[i] = grid[2*(ij+i)];
				ncvarput (cdfid, z_id, start, edge, (void *)tmp);
			}
			else
				ncvarput (cdfid, z_id, start, edge, (void *)&grid[ij]);
		}
	}
	ncclose (cdfid);
	
	if (complex || geo) free ((char *)tmp);
	
	return (ncerr);
}


/*
 * stolen from GMT3.0/src/gmt_support.c
 */

char *memory (prev_addr, n, size, progname)
char *prev_addr, *progname;
int n, size; {
	char *tmp;

	if (n == 0) return((char *) NULL); /* Take care of n = 0 */
	
	if (prev_addr) {
		if ((tmp = realloc ((char *) prev_addr, (unsigned) (n * size))) == CNULL) {
			fprintf (stderr, "GMT Fatal Error: %s could not reallocate more memory, n = %d\n", progname, n);
			exit (-1);
		}
	}
	else {
		if ((tmp = calloc ((unsigned) n, (unsigned) size)) == CNULL) {
			fprintf (stderr, "GMT Fatal Error: %s could not allocate memory, n = %d\n", progname, n);
			exit (-1);
		}
	}
	return (tmp);
}
